module vga9 (clk,r,g,b,hs,vs,vgaclk,       // vga interface
             addr,datain,dataout,wr,rd,ready,      // cpu interface
             sram_addr,sram_ce_n,sram_dq,sram_lb_n,sram_oe_n,sram_ub_n,sram_we_n,
				 debug);// sram interface); 
				 // 1440x900 160x50 char with border text mode vga driver for DE2-115

// ver 0.15 alpha --- 2014.09.03
// (c) Piotr Kardasz pik33@we.pb.edu.pl
// This is free software, license: LGPL v.2.0 or newer

output reg [7:0] r,g,b;
output reg hs,vs;
output wire vgaclk; 
input wire clk;  //input clock=2x vgaclk
output wire ready;

output reg[7:0] debug;

reg [3:0] e;
reg f,oldwr,oldwr2;
reg [7:0] brr,brb,brg;
reg [19:0] caddr;
reg [10:0] lcnt;
reg [0:2] lstate,pstate; // 00 vis, 01 fp, 10 sync, 11 bp 100 border
reg [11:0] pcnt;
reg [31:0] fcnt; //frame counter

input wire [20:0] addr;
input wire [31:0] datain;
output reg [31:0] dataout;
input wire wr,rd;
output reg [19:0] sram_addr;
inout wire [15:0] sram_dq;
output wire sram_ce_n;
output reg sram_lb_n;
output wire sram_oe_n;
output reg sram_ub_n;
output reg sram_we_n;
reg [15:0] sramdqout;
reg[31:0] datain2,datain3;
reg [20:0] addr2,sram_addr2;
reg[3:0] pcnt_mod,pcnt_mod2;
reg[7:0] pcnt_char;
reg[15:0] pixels3, pixels2;
reg[31:0] pixels;
reg mode,wr2,wr3,rd2,rd3,oldrd;
reg[3:0]wstate;


(* ramstyle = "M9K" *) reg[31:0] regs [32]; // internal registers
(* ramstyle = "M9K" *) reg [7:0] fontdef[2048];  // character generator

// 0 - display state: {vsync,hsync,frame counter}
// 1 - display start address
// 2 - mode
// 15 - border color 
// 16..31: pallette

initial
  begin
  $readmemh("st4font.hex",fontdef);
  end

  
assign vgaclk=clk;
assign sram_ce_n=0;
assign sram_oe_n=0;
 

// ------------------------------------------- 
// bidir assignment
// ------------------------------------------- 

assign sram_dq = (sram_we_n==0) ? sramdqout : 16'bz;

//---------------------------------------------  
// debug  
//---------------------------------------------  
  
always@(posedge clk) 
 
begin 



debug[3:0]<=wstate;
debug[7]<=wr;
debug[6]<=wr3;
debug[5]<=rd;
debug[4]<=rd3;
end



//---------------------------------------------
// main loop
//--------------------------------------------- 
 
always@(posedge clk)

  begin
  oldwr<=wr;
  if ((oldwr==0) && (wr==1)) wr2<=1; else wr2<=0;
  wr3<=wr2;
  oldrd<=rd;
  if ((oldrd==0) && (rd==1)) rd2<=1; else rd2<=0;
  rd3<=rd2;
  pcnt_mod2<=pcnt_mod-1;
  
// ------------------------------------------- 
// sram rw state machine
// -------------------------------------------
  
// new state calculation
  case (wstate) 
  
    0: begin
	   addr2<=addr;
		if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==0)) wstate<=1;
		else if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==1)) wstate<=3;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==0)) wstate<=4;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==1)) wstate<=5;
		else wstate<=0;
     end
	  
	1: begin  
	 
		if ((pcnt_mod2==0) || (pcnt_mod2==1)) wstate<=6; else wstate<=1; 
		datain2=datain;
		sramdqout<=datain2[31:16];
		sram_addr2<={addr,1'b0};
      end 

   6: begin		
	
		if (pcnt_mod2==4) wstate<=7; else wstate<=6; 
  	   sram_addr<=sram_addr2;
	   sram_we_n<=0;    
      sram_lb_n<=0;
      sram_ub_n<=0;  
      end

	7: begin

		if (pcnt_mod2==5) wstate<=8; else wstate<=7;
      sram_lb_n<=1;
      sram_ub_n<=1;	 
   	sram_addr<=sram_addr2+1;
  	   sramdqout<=datain2[15:0];
	   end
  
   8: begin
  
      if (pcnt_mod2==7) wstate<=9; else wstate<=8;
      sram_lb_n<=0;
      sram_ub_n<=0;	 
  	   end	  
  
   9: begin
  
		if (pcnt_mod2==8) wstate<=0; else wstate<=9;
   	sram_we_n<=1;    
      sram_lb_n<=1;
      sram_ub_n<=1;  
      end

	4:	begin
	 
		if ((pcnt_mod2==0) || (pcnt_mod2==1)) wstate<=10; else wstate<=4;
		sram_addr2<={addr,1'b0};
      end		
		
	10:begin	
				
		if (pcnt_mod2==3) wstate<=11; else wstate<=10;
  	   sram_addr<=sram_addr2;
	   sram_we_n<=1;    
      sram_lb_n<=0;
      sram_ub_n<=0;  
      end		
		
	11:begin
	
		if (pcnt_mod2==4) wstate<=12; else wstate<=11;
		sram_addr<=sram_addr2+1;
		dataout[31:16]<=sram_dq;
	   end		
		
	12:begin	
		
		if (pcnt_mod2==7) wstate<=13; else wstate<=12;
      end
		
   13:begin
		if (pcnt_mod2==8) wstate<=0; else wstate<=13;
	 	dataout[15:0]<=sram_dq; 
		end
		 
   3: begin  
      wstate<=0;
      if (addr[20]==0) regs[addr2[4:0]]<=datain; else fontdef[addr2[10:0]]<=datain[7:0];
   	end
		
	5: begin	

  	   wstate<=0;
	   if (addr[20]==0) dataout<=regs[addr2[4:0]]; else dataout<={24'b0,fontdef[addr2[10:0]]};
	   end
	 
   default: wstate<=0;
	
	endcase
	
//----- end of sram interface state machine

//----- vga 
	 
  if (pcnt_mod==9)	 	 
    begin
	 sram_addr<=caddr;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 
	 
  if (pcnt_mod==11)	 	 
    begin
	 sram_addr<=caddr+1;
	 pixels2<=sram_dq;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 

  if (pcnt_mod==13)	 	 
    begin
	 pixels3<=sram_dq;

	 end 	 
	 
  if (pcnt_mod==0) pixels={pixels2,pixels3};

  	 pcnt<=pcnt+1;
    pcnt_mod<=(pcnt+1)&15;
	 pcnt_char<=(pcnt-304)>>4;
    // line and pixel counters & state determination
	 if (pcnt==239) begin pstate<=4; regs[0]<={vs,20'd0,lcnt}; end  // left border
	 if (pcnt==319) pstate<=0; // display
	 if (pcnt==1599) pstate<=4; // right border
	 if (pcnt==1679) begin pstate<=1; regs[0]<={vs,1'b1,19'd0,lcnt}; end // front porch
	 if (pcnt==1759) begin pstate<=2; hs<=0; end// hsync
	 if (pcnt==1911)           // back porch 
	   begin 
		pstate<=3; 
      hs<=1;
		pcnt<=8;
		lcnt<=lcnt+1; 
		mode<=regs[2][0];
		if (lcnt==27) lstate<=4;   // top border
		if (lcnt==77) lstate<=0;   // display
		if (lcnt==877) lstate<=4;  // bottom border
		if (lcnt==927) lstate<=1;  // front porch v
		if (lcnt==928) begin lstate<=2; vs<=1; end // vsync 
		if (lcnt==931) begin lstate<=3; vs<=0; lcnt<=0; fcnt<=fcnt+1; end // back porch v
		end
 
    caddr=((lcnt-78)>>mode)*160+(pcnt_char<<1)+(regs[1]<<1);
	 

	 if (pstate==1)
	   begin
		brr<=regs[15][7:0];
		brg<=regs[15][15:8];
		brb<=regs[15][23:16];
      end 
    
    if (((lstate==4)&&((pstate&3)==0)) || ((pstate==4)&&(lstate==0))) // borders
      begin 
      r<=brr; g<=brg; b<=brb;	

      end		 
	 
    else if (lstate==0 && pstate==0) // active display
      begin


		if (mode==0)
        case (pcnt_mod)
          0:e=pixels[31:30];
	  	    1:e=pixels[29:28];
		    2:e=pixels[27:26];
		    3:e=pixels[25:24];
		    4:e=pixels[23:22];
		    5:e=pixels[21:20];
		    6:e=pixels[19:18];
		    7:e=pixels[17:16];		            
			 8:e=pixels[15:14];
	  	    9:e=pixels[13:12];
		    10:e=pixels[11:10];
		    11:e=pixels[9:8];
		    12:e=pixels[7:6];
		    13:e=pixels[5:4];
		    14:e=pixels[3:2];
		    15:e=pixels[1:0];		  
		  endcase
		else  
		   case (pcnt_mod)
  		   0,1:e=pixels[31:28];
		   2,3:e=pixels[27:24];
		   4,5:e=pixels[23:20];
		   6,7:e=pixels[19:16];  		   
			8,9:e=pixels[15:12];
		   10,11:e=pixels[11:8];
		   12,13:e=pixels[7:4];
		   14,15:e=pixels[3:0];

      endcase
        case (e)
		  0: begin r<=regs[16][7:0]; g<=regs[16][15:8]; b<=regs[16][23:16]; end
		  1: begin r<=regs[17][7:0]; g<=regs[17][15:8]; b<=regs[17][23:16]; end
		  2: begin r<=regs[18][7:0]; g<=regs[18][15:8]; b<=regs[18][23:16]; end
		  3: begin r<=regs[19][7:0]; g<=regs[19][15:8]; b<=regs[19][23:16]; end
		  4: begin r<=regs[20][7:0]; g<=regs[20][15:8]; b<=regs[20][23:16]; end
		  5: begin r<=regs[21][7:0]; g<=regs[21][15:8]; b<=regs[21][23:16]; end
		  6: begin r<=regs[22][7:0]; g<=regs[22][15:8]; b<=regs[22][23:16]; end
		  7: begin r<=regs[23][7:0]; g<=regs[23][15:8]; b<=regs[23][23:16]; end
		  8: begin r<=regs[24][7:0]; g<=regs[24][15:8]; b<=regs[24][23:16]; end
		  9: begin r<=regs[25][7:0]; g<=regs[25][15:8]; b<=regs[25][23:16]; end
		  10: begin r<=regs[26][7:0]; g<=regs[26][15:8]; b<=regs[26][23:16]; end
		  11: begin r<=regs[27][7:0]; g<=regs[27][15:8]; b<=regs[27][23:16]; end
		  12: begin r<=regs[28][7:0]; g<=regs[28][15:8]; b<=regs[28][23:16]; end
		  13: begin r<=regs[29][7:0]; g<=regs[29][15:8]; b<=regs[29][23:16]; end
		  14: begin r<=regs[30][7:0]; g<=regs[30][15:8]; b<=regs[30][23:16]; end
		  15: begin r<=regs[31][7:0]; g<=regs[31][15:8]; b<=regs[31][23:16]; end


  	     endcase

		end
	 
	 else begin r<=0; g<=0; b<=0; end		

  end
 
endmodule 